# moflextab.py. This file automatically created by PLY (version 2.3). Don't edit!
_lextokens    = {'DT_UINT8': None, 'DISABLEOVERRIDE': None, 'PRAGMA': None, 'DT_SINT64': None, 'octalValue': None, 'hexValue': None, 'stringValue': None, 'decimalValue': None, 'NULL': None, 'PARAMETER': None, 'charValue': None, 'DT_DATETIME': None, 'REFERENCE': None, 'RESTRICTED': None, 'TOSUBCLASS': None, 'DT_REAL64': None, 'TRANSLATABLE': None, 'ENABLEOVERRIDE': None, 'DT_UINT32': None, 'ASSOCIATION': None, 'SCHEMA': None, 'DT_BOOL': None, 'floatValue': None, 'METHOD': None, 'DT_SINT16': None, 'AS': None, 'DT_SINT32': None, 'binaryValue': None, 'DT_STR': None, 'SCOPE': None, 'FLAVOR': None, 'DT_CHAR16': None, 'TRUE': None, 'DT_REAL32': None, 'DT_SINT8': None, 'FALSE': None, 'QUALIFIER': None, 'DT_UINT64': None, 'OF': None, 'REF': None, 'CLASS': None, 'INSTANCE': None, 'IDENTIFIER': None, 'PROPERTY': None, 'INDICATION': None, 'DT_UINT16': None, 'ANY': None}
_lexreflags   = 0
_lexliterals  = '#(){};[],$:='
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_COMMENT>//.*)|(?P<t_MCOMMENT>/\\*(.|\\n)*?\\*/)|(?P<t_IDENTIFIER>([a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))([0-9a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))*)|(?P<t_newline>\\n+)|(?P<t_stringValue>"([^"\\\\\\n\\r]|([\\\\](([bfnrt\'"\\\\])|(x[0-9a-fA-F]{1,4}))))*")|(?P<t_floatValue>[+-]?[0-9]*\\.[0-9]+([eE][+-]?[0-9]+)?)|(?P<t_hexValue>[+-]?0[xX][0-9a-fA-F]+)|(?P<t_decimalValue>[+-]?([1-9][0-9]*|0))|(?P<t_binaryValue>[+-]?[01]+[bB])|(?P<t_octalValue>[+-]?0[0-7]+)', [None, ('t_COMMENT', 'COMMENT'), ('t_MCOMMENT', 'MCOMMENT'), None, ('t_IDENTIFIER', 'IDENTIFIER'), None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, ('t_newline', 'newline'), (None, 'stringValue'), None, None, None, None, None, (None, 'floatValue'), None, (None, 'hexValue'), (None, 'decimalValue'), None, (None, 'binaryValue'), (None, 'octalValue')])]}
_lexstateignore = {'INITIAL': ' \r\t'}
_lexstateerrorf = {'INITIAL': 't_error'}
